/***************************************************************************
 *            ut_text.h
 *
 *  Tue Oct  5 11:28:11 2004
 *  Copyright  2004  Antoine Calando
 *  antoine@alliancemca.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*!
 * \file ut_text.h
 * \author Antoine Calando (antoine@alliancemca.net)
 */

#ifndef _UT_TEXT_H_
#define _UT_TEXT_H_



/***************************************************************************/
/*!
 * \brief Flags that control the processing of a text.
 * 
 * They are set by the user to tune the way the text will be analysed
 * and converted. Some of them are unimplemented (UT_F_REFERENCE_EXT_CHAR, always true).
 *
 */

typedef enum UtTextFlags {
	//UT_F_UNSET = 0,
	UT_F_READ_ONLY_DATA = 1<<0,		//!< Do not modify the text.
	UT_F_TRANSFORM_EOL = 1<<1,		//!< Replace EOL by null character to simplify the
									//!< processing.
	UT_F_ADD_FINAL_EOL = 1<<2,		//!< Add a final EOL to the text if the last line is not empty.
	UT_F_REFERENCE_EXT_CHAR = 1<<3,	//!< Register the lines that contains extended characters (unimplemented, always true).
	UT_F_REMOVE_ILLEGAL_CHAR = 1<<4,	//!< Remove control characters (except CR, LF and TAB).
	UT_F_FORCE_BINARY = 1<<5,		//!< Forece processing of the file even if it is detected as binary data.
	//UT_F_IDENTIFY_EOL = 1<<5,
	//UT_F_IDENTIFY_CHARSET = 1<<6,
	UT_F_DEFAULT = UT_F_TRANSFORM_EOL | UT_F_REFERENCE_EXT_CHAR | UT_F_REMOVE_ILLEGAL_CHAR | UT_F_ADD_FINAL_EOL
} UtTextFlags;


/***************************************************************************/
/*!
 * \brief Contains evaluation of a charset.
 *
 * An array of this structure is instanciated in UtText and holds the result of
 * the evaluation of each charset. The charset which get the best rating will
 * be choosed for the conversion.
 */

typedef struct UtCharsetEval {
	long rating; 	//!< Mark attributed to the charset depending on the text
	ulong checksum;	//!< Checksum of each extended character in the text. Used to find equivalent charsets.
} UtCharsetEval;


/***************************************************************************/
/*!
 * \brief Refers to a line with extended characters.
 *
 * This structure refers to a line with extended characters.
 * The list of lines with extended characters is filtered to exclude lines
 * with same characters and is stocked in a linked list accessible from UtText.
 */

typedef struct UtExtCharLine {
	char * line_p;		//!< Pointer to the beginning line.
	ulong line_i;		//!< Number of the line.
	ulong ext_char_n;	//!< Number of extended characters in the line.
	struct UtExtCharLine * next;	//!< Pointer to the next struture. NULL if last.
} UtExtCharLine;


/***************************************************************************/
/*!
 * \brief Types of End-of-line characters.
 *
 * Different types are CRLF (DOS/Windows), LF (Unix), CR (Mac). The types CRLF_CR and
 * CRLF_LF exists in some CSV databases : entries are ended with CRLF, but some fields
 * may contains LF or CR alone to indicate a "carriage return" in the field.
 * CR is the character 0xD, LF is 0xA.
 *
 * \note EC le cas du LFCR n'est pas pris en compte (cela n'existe pas ?)
 *       AC Si! je ne l'ai pas rencontré, mais il faudrait le rajouter...
 *       (en fait il faudrait même modifier pas mal de trucs dans la reconnaissance
 *       de fins de ligne)
 */

typedef enum UtEolType {
	//UT_EOL_NO_CHANGE = -2,
	UT_EOL_UNSET=-1, 
	UT_EOL_CRLF,
	UT_EOL_CRLF_CR,
	UT_EOL_CRLF_LF,
	UT_EOL_CR,
	UT_EOL_LF,
	UT_EOL_NON_STD
} UtEolType;

#ifdef _UTRAC_C_
/*!
 *\brief Names of eol type.
 */
const char * UT_EOL_NAME[] = { "CRLF", "CRLF_CR", "CRLF_LF","CR", "LF", "NON_STD" };
#else
extern const char * UT_EOL_NAME [];
#endif

typedef short UtCharsetIndex;



/***************************************************************************/
/*!
 * \brief Contains all the information about a text and its processing.
 *
 * This structure is created by ut_init_text() and destroyed by ut_free_text(). It is used
 * to pass different arguments to ut_process_text(), and to stock information about the
 * text all along its processing.
 */

typedef struct UtText {
	const char * filename; 			//!< Name of the file. If NULL, stdin is used. Set by User.
	char * data; 					//!< Pointer to the beginning of the text. It is finished by a null character. Set by user or Utrac.
	ulong size;						//!< Size of the text, without the terminating null character. Set by user or Utrac.
	ulong line_n;					//!< Number of lines in the text. Set by Utrac.
	//ulong line2_n;
	UtTextFlags flags;				//!< Flags that control the processing of the text. Set by user.
	//UtProgress state;

	ulong distribution [256];		//!< Frequency distribution of the text. Set by Utrac.
	UtExtCharLine * ext_char;		//!< Linked list of lines containing extended characters. Set by Utrac.
	UtCharsetEval * evaluation;		//!< Array containg evaluation of each charset. Set by Utrac.
	
	UtEolType reco_eol;				//!< EOL type recognized by Utrac.
	UtCharsetIndex reco_charset;	//!< Charset recognized by Utrac.

	UtEolType src_eol;				//!< EOL Type used as source for conversion. If UT_EOL_UNSET, #reco_eol is used. 
									//!< (unimplemented, recognized value is always used). Set by user or Utrac.
	UtCharsetIndex src_charset;		//!< Charset used as source for conversion. If UT_UNSET, #reco_charset is used. Set by user or Utrac.

	UtEolType dst_eol;				//!< EOL Type used as destination for conversion. Set by user or Utrac.
	UtCharsetIndex dst_charset;		//!< EOL Type used as destination for conversion. Set by user or Utrac.

	char skip_char;					//!< Character to skip during conversion. A variable is used rather than the constant 
									//!< UT_SKIP_CHAR, since the text can already already contains UT_SKIP_CHAR value if
									//!< UT_F_REMOVE_ILLEGAL_CHAR is not set. Set by user.
	
	ulong nomapping_char;			//!< Character used if a character conversion error occurs. No character inserted if zero. Set by user.
	
	int (*progress_function)(float);//!< User function called during processing to update a progress bar. Can be NULL.
									//!< The function is called regularly (see UT_LOAD_TEXT_STEP and UT_PROCESS_TEXT_STEP),
									//!< with the float argument betwwen 0.0 and 1.0, indicating the part of the job done.
									//!< The function is called only once with 0.0 (the first time) and once with 1.0 (the
									//!< last time), so these values can be checked to do initialisation and cleanup. 
									//!< If this function return 0, the processing is interrupted. Set by user.
	float progress_done;			//!< Part of the process already done. Value included between 0.0 and 1.0. Set by Utrac.
	int progress_todo;				//!< Number of passes to do before end of the process. Set by Utrac.

} UtText;

#endif //_UT_TEXT_H_
