/***************************************************************************
 *            utrac.c
 *
 *  Tue Oct  5 11:29:59 2004
 *  Copyright  2004  Antoine Calando
 *  antoine@alliancemca.net
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
/*!
 * \file utrac.c
 * \author Antoine Calando (antoine@alliancemca.net)
 * \brief Public API for using Utrac.
 */
 
#define _UTRAC_C_

#include <stdlib.h>
#include <stdio.h>
#include "utrac.h"

#undef UT_DEBUG
#define UT_DEBUG 3
#include "debug.h"


/***************************************************************************/
/*!
 * \brief Initialize the Utrac library.
 *
 * This function must be called before any other Utrac function. It allocates an UtSession struture
 * that is accessible by the ut_session pointer, initalizes it, loads charsets data, and sets
 * default	language, charset and end of line type. The memory used is about 630kb for 47 charsets
 * loaded.
 *	
 * \note On Unix systems, LANG, LC_ALL and LC_TYPE are parsed to find default language and
 * charsets (ISO-8859-1 is used else), and	eol type is set to LF.
 *
 * \return UT_OK on success, an error code on failure.
 */
UtCode ut_init () {
	
	if (ut_session) return UT_ALREADY_INITIALISED_ERROR;
		
	ut_session = (UtSession*) malloc (sizeof(UtSession));
	if (!ut_session) return UT_MALLOC_ERROR;

	//ut_session->flags = UT_F_UNSET; //flags_in
	ut_session->charset = NULL;
	ut_session->charset_n = 0;
	ut_session->language.name = NULL;
	ut_session->language.code = NULL;
	ut_session->language.n = 0;
	ut_session->language.n_max = 0;
	ut_session->system.name = NULL;
	ut_session->system.code = NULL;
	ut_session->system.n = 0;
	ut_session->system.n_max = 0;
	ut_session->error_string[0] = '\0';
	ut_session->charset_default = UT_UNSET;
	ut_session->eol_default = UT_EOL_UNSET;

	//load charsets data
	UT_TRY (ut_load_charsets ())

	//find default language, charset, eol type on the system
	#ifdef linux
	//should we use nl_langinfo()? (discovered later...)
	int i;
	ut_session->language_default = UT_UNSET; //language_default_in
	ut_session->eol_default = UT_EOL_LF;
	ut_session->charset_default = UT_UNSET;
	
	char * def_enc = getenv ("LC_CTYPE");
	if (!def_enc) def_enc = getenv ("LC_ALL");
	if (!def_enc) def_enc = getenv ("LANG");
	if (def_enc) {
		if (def_enc[2]=='_' || def_enc[2]=='.' || def_enc[2]==0) {
			for (i=0; i<ut_session->language.n; i++) {
				if ((def_enc[1]-'a'+'A'== (ut_session->language.code[i]>>8 &0xFF))
					&& (def_enc[0]-'a'+'A'== (ut_session->language.code[i] &0xFF)) ) {
					ut_session->language_default = i;
					break;
				}
			} //for
		}
		if (def_enc[2]=='.') def_enc +=3;
		if (def_enc[2]=='_' && def_enc[5]=='.') def_enc +=6;
		for (i=0; i<ut_session->charset_n; i++) 
			if (ut_str_fuzzy_cmp (def_enc, ut_session->charset[i].name,'@')) break;
		if (i!=ut_session->charset_n) ut_session->charset_default = i;
	}
	
	if (ut_session->charset_default == UT_UNSET) {
		for (i=0; i<ut_session->charset_n; i++) 
			if (ut_str_fuzzy_cmp (UT_DEFAULT_ENCODING_UNIX, ut_session->charset[i].name,0)) break;
		if (i==ut_session->charset_n) {
			DBG1 ("*** No default charset ***")
		}
		else ut_session->charset_default = i;
	}
	#else
	DBG3 ("pas unix!")
	#endif

	#if UT_DEBUG == 3
	if (ut_session->language_default != UT_UNSET)
		DBG3 ("lang: %s" , ut_session->language.name[ut_session->language_default])
	if (ut_session->charset_default != UT_UNSET)
		DBG3 ("charset: %s", ut_session->charset[ut_session->charset_default].name)
	if (ut_session->eol_default != UT_EOL_UNSET)
		DBG3 ("eol: %s", UT_EOL_NAME [ut_session->eol_default])
	#endif
	
	return UT_OK;
}

/***************************************************************************/
/*!
 * \brief Free ressources allocated during initialization of Utrac.
 *
 * This function frees the structure allocated in ut_session
 * by ut_init(). It must be the last Utrac function called.
 *
 * \return UT_OK on success, an error code on failure.
 */

UtCode ut_finish () {
	
	if (!ut_session) return UT_ERROR;
	
	int i; for(i=0; i<ut_session->charset_n; i++) {
		free(ut_session->charset[i].name);
		free(ut_session->charset[i].alias);
		free(ut_session->charset[i].common_name);
		free(ut_session->charset[i].comment);
		free(ut_session->charset[i].unicode);
		free(ut_session->charset[i].char_type);
		free(ut_session->charset[i].language);
		free(ut_session->charset[i].system);
	}
	free (ut_session->charset);
	
	for (i=0; i<ut_session->language.n; i++) 
		free (ut_session->language.name[i]);
	free (ut_session->language.name);
	free (ut_session->language.code);

	for (i=0; i<ut_session->system.n; i++) 
		free (ut_session->system.name[i]);
	free (ut_session->system.name);
	free (ut_session->system.code);

	free(ut_session);
	ut_session = NULL;
	
	return UT_OK;
}


/***************************************************************************/
/*!
 * \brief Allocates and initalizes an UtText structure.
 *
 * \return A pointer to the allocated structure, or NULL if the allocation failed.
 *
 * \note Some members of this struture must be set before it is used.
 */

UtText * ut_init_text () {
	ASSERT (ut_session)
	UtText* new_text = (UtText*) malloc (sizeof(UtText));
	if (!new_text) return NULL;
		
	new_text->filename = NULL; 
	new_text->data = NULL;
	new_text->size = 0;
	new_text->line_n = 0;
	int i; for (i=0; i<0x100; i++) new_text->distribution [i] = 0;
	new_text->flags = UT_F_DEFAULT;

	new_text->ext_char = NULL;
	new_text->evaluation = NULL;
	
	new_text->reco_eol = UT_EOL_UNSET;
	new_text->reco_charset = UT_UNSET;
	new_text->src_eol = UT_EOL_UNSET;
	new_text->src_charset = UT_UNSET;
	
	new_text->dst_eol = UT_EOL_UNSET;
	new_text->dst_charset = UT_UNSET;

	new_text->skip_char = UT_SKIP_CHAR;
	new_text->nomapping_char = 0x7F;
	
	new_text->progress_done = 0.0;
	new_text->progress_todo = 0;
	new_text->progress_function = NULL;
	
	return new_text;
}

/***************************************************************************/
/*!
 * \brief Free an UtText structure.
 *
 * \param text pointer to the structure to free.
 *
 * \note EC avant de faire un free on vérifie que le pointeur n'est pas NULL et
 *       on le met à NULL après le free.
 *       AC pas OK : free(NULL) est tout a fait légal et les pointeurs à mettre à
 *       NULL sont dans une structure qui va être désalouée
 */

void ut_free_text (UtText *text) {
	//free(text->filename);
	//filename is not freed because it is set by user. 
	free(text->data);
	while (text->ext_char) {
		UtExtCharLine * tmp = text->ext_char;
		text->ext_char = text->ext_char->next;
		free (tmp);
	}
	if (text->evaluation) free(text->evaluation);

	free(text);	
	
}

/***************************************************************************/
/*!
 * \brief Recognize charset and EOL type of a text, and eventually convert it.
 * 
 * This function take an UtText structure as a parameter and do severeal tasks :
 * -# it loads the file (or read the standard input),
 * -# it calculate the frequency distribution of each byte in the file
 *    (UtText::distribution), checks if the file is binary data or text,
 *    checks if it is ASCII or UTF-8,
 * -# it recognize the EOL type, and replace each EOL by null character to make
 *    further processing of the file easier (this feature can be disbled).
 * -# if the charset has not been determined earlier as ASCII or UTF-8, it tries
 *    to detect which known charset fit the best to the text.
 * -# it eventually convert the text, replacing EOL and extended character by
 *    those corresponding to the selection of the user and/or the result of the recogntion.
 *
 * \param text Text to recognize and eventually convert. Some members must be set
 *        before calling this function, but some other are optionnal. Members that
 *        select the input text are:
 * 	      - UtText::data: Pointer to the text to process (which must be null terminated).
 *          If NULL, UtText::filename is used.
 *        - UtText::filename: Path to the file containing the text to process, which will
 *          be loaded if . If NULL, standard input is read.
 * 	      - UtText::size: If UtText::data is set, this member can also be set to indicate
 *          the size of the text, if null, the first null character will determine the
 *          end of the text.
 * 
 * Members that modifies the recognition or the conversion are:
 * 	- UtText::flags: Flags to customize the processing and the modification of the text.
 *    Set intially to UT_F_DEFAULT.
 * 	- UtText::src_eol and UtText::src_charset: EOL type and charset of the text used as
 *    source for the conversion. If unset, the values taken are those recognized automatically.
 * 	- UtText::dst_eol and UtText::dst_charset: EOL type and charset of the text resulting of
 *    the conversion. If unset, the values taken are those by default found by ut_init().
 * 	- UtText::nomapping_char: Character inserted during the conversion each time an error occurs.
 * 
 * Misc member:
 * 	- UtText::progress_function: Custom function provided by the user to refresh a progress bar.
 *
 * \param convert If true, conversion is effectued after recognition.
 *	
 * \return UT_OK on success, error code on failure (see UtCode).
 */


UtCode ut_process_text (UtText * text, bool convert) {
	
	if (!text) return UT_BAD_PARAMETER_ERROR;

	//init text->progress_*
	if (text->progress_function)  {
		if (text->progress_done) {
			text->progress_done = 0.0;
			text->progress_todo = 0;
		}
		
		bool already_init;
		if (text->progress_todo) already_init = true;
		else already_init = false;
	
		text->progress_todo += 4;
		if (convert) text->progress_todo++;
		if (!already_init) ut_update_progress (text, 0, true);
	}

	//load file if none provided
	if (!text->data) {
		if (text->filename) {
			UT_TRY (ut_get_file_size (text->filename, &text->size))
			UT_TRY (ut_load_text (text))
		} else {
			UT_TRY (ut_load_text_stdin (text))
		}			
		
	}
	
	//update progress_*
	if (text->progress_function) {
		text->progress_done+= (1-text->progress_done)/text->progress_todo;
		text->progress_todo--;
	}
	
	//checks if flags are correct
	if (text->flags & UT_F_READ_ONLY_DATA) {
		if (text->flags & (UT_F_TRANSFORM_EOL|UT_F_ADD_FINAL_EOL|UT_F_REMOVE_ILLEGAL_CHAR))
			return UT_BAD_FLAGS_ERROR;
	}
	
	// todo: modify skip_char to be sure that it doesn't interfer with the text that use
	// the same value.
	text->skip_char = UT_SKIP_CHAR;
	
	//first pass to detect if file is binary, ASCII,  or UTF-8
	int rcode = ut_distrib_utf_pass (text);
	if (rcode == UT_BINARY_DATA_ERROR) {
		if ( !(text->flags & UT_F_FORCE_BINARY)) return rcode;
	} else if ( rcode != UT_OK) return rcode;
	
	//update progress_*
	if (text->progress_function) {
		text->progress_done+= (1-text->progress_done)/text->progress_todo;
		if (text->reco_charset==UT_UNSET)
			text->progress_todo--;
		else text->progress_todo-=2;
	}

	//second pass to recognize end of file	
	UT_TRY ( ut_eol_pass (text) )

	//update progress_*
	if (text->progress_function) {
		text->progress_done+= (1-text->progress_done)/text->progress_todo;
		text->progress_todo--;
	}

	//if charset still not recognized, third pass to recognize extended ascii charset	
	if (text->reco_charset==UT_UNSET) UT_TRY ( ut_xascii_pass (text) )

	//if src_* or dst_* unset, set them with default values
	if (text->src_charset == UT_UNSET) text->src_charset = text->reco_charset;
	if (text->src_eol == UT_EOL_UNSET) text->src_eol = text->reco_eol;
	if (text->dst_charset==UT_UNSET) text->dst_charset = ut_session->charset_default;
	if (text->dst_eol==UT_EOL_UNSET) text->dst_eol = ut_session->eol_default;

	if (convert) {	
		//update progress_*
		if (text->progress_function) {
			text->progress_done+= (1-text->progress_done)/text->progress_todo;
			text->progress_todo--;
		}
		//fourth pass
		UT_TRY ( ut_conversion_pass (text) )
	}
	
	//update progress_*
	if (text->progress_function) {
		text->progress_done+= (1-text->progress_done)/text->progress_todo;
		if (!--text->progress_todo)	ut_update_progress (text, 0, true);
	}
	
	return UT_OK;
}
